package CH.ifa.draw.contrib.xcomp;

import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import java.awt.*;
import java.util.*;
import java.io.*;

/**
 * The base class for implementing list-like layout strategies.
 * AbstractFlowLayout maintains an ordered vector of components that
 * belong to the layout. The ordering is independent
 * of the order of the <em>display list</em> maintained by the container
 * (CompositeFigure) that determines the drawing order of components.
 */ 

public abstract class AbstractFlowLayout extends BagLayout {
	
  transient protected Vector fArrangement = null;

  protected Vector fComponents;

	Dimension emptySize;
	
	public AbstractFlowLayout() {
		fComponents = new Vector();
	}
	
	/**
	 * Force rearrangement.
	 */
	public void invalidateLayout() {
		//System.err.println("abstract flow layout invalidated");
		fArrangement = null;
		attachAligners();
	}
	
	
	public Dimension getEmptySize() {
		return emptySize;
	}
	
	public void  setEmptySize(Dimension dim) {
		emptySize = dim;
	}
	
		/**
	 * Force rearrangement.
	 */
	public void invalidateLayout(Figure component) {
		//System.err.println("abstract flow layout invalidated with component");
		//moveIn((CompositeFigure)component.parent(),component);
		invalidateLayout();	
	}

	/**
	 * Adds a component to the layout. Uses findPos to find the correct
	 * place for the component. Moves the component so that its origin
	 * is wihtin its container's display box, if necessary.
	 */
	public void addLayoutComponent(Figure component) {
		if (!fComponents.contains(component)) {
			fComponents.insertElementAt(component,findPos(component));
			//grow((CompositeFigure)component.parent(),component.displayBox());
			//moveIn((CompositeFigure)component.parent(),component);
		}
	}	
	/**
	 * Does nothing. Bag layout does not need to store the components.
	 */
	public void removeLayoutComponent(Figure component) {
			if (fComponents.contains(component)) {
				fComponents.removeElement(component);
			}
	}
	
	/**
	 * Does nothing. Bag layout does not need to store the components.
	 */
	public void replaceLayoutComponent(Figure component, Figure replacement) {
		if (fComponents.contains(component)) {
			fComponents.setElementAt(
				replacement,
				fComponents.indexOf(component)
			);
				attachAligners();
		}			
	}

	/**
	 * Returns the components currently stored by this layout. The order of
	 * components may be significant to the layout.
	 */
	public Vector getComponents() {
		return (Vector)fComponents.clone();
	}

	// make this a property
	protected Dimension emptySize() {
		if (emptySize == null)
			emptySize = new Dimension(10,10);
		return emptySize;
	}

  /**
   * Gets the current arrengement of subfigures. Computes the arrengement
   * if necessary.
   */
  protected Vector getArrengement(CompositeFigure container) {
    //System.err.println("FigureHolder#getArrengement()");
    if (fArrangement == null) {
      fArrangement = new Vector(); //always compute this??
      arrange(container);
    }
    return fArrangement;
  }
  /**
   * Return the holdees that are stationary, that is, not in transit. 
   * Includes also figures
   * that are selected to be dragged but not dragged yet. 
   */
  protected Vector getStationaryFigures() {
    Vector result = new Vector();
    Enumeration holdee = fComponents.elements();
    while (holdee.hasMoreElements()) {
      Figure figure = (Figure)holdee.nextElement();
      // might be a multiple selection - figures are selected to be dragged
      // but not dragged yet
      if (!figure.hasMoved()) //!figure.inTransit())
        result.addElement(figure);
    }
    return result;
  }

	/**
	 * Lays out the Sovereign figures that are contained in the container.
	 * Sovereigns are placed independently of the container but they need
	 * the parent to initiate the lay out process. 
	 */
	protected void layoutSovereigns(CompositeFigure container) {
	//	System.err.println("<><><><>");
    Enumeration holdee = container.figures();
    while (holdee.hasMoreElements()) {
      Figure figure = (Figure)holdee.nextElement();
      // might be a multiple selection - figures are selected to be dragged
      // but not dragged yet
      //System.err.println("f: "+figure+" is sovereign: "+(figure instanceof Sovereign));
      if (!figure.hasMoved() && (figure instanceof Sovereign) && !figure.isValid()) //!figure.inTransit())
        accommodate(container,figure);
    }		
		//System.err.println("<><><><>");
	}

  protected void attachAligners() {
  	//System.err.println("@t@s");
    /*Enumeration figure = getStationaryFigures().elements();
    if (figure.hasMoreElements()) {
      Figure childF = (Figure)figure.nextElement();
      //moveIn((CompositeFigure)childF.parent(),childF);
      childF.setAligner(
      	getTopElementAligner((CompositeFigure)childF.parent(),childF)
      );
      Figure prev = childF;
		  while (figure.hasMoreElements()) {
        childF = (Figure)figure.nextElement();
		    //moveIn((CompositeFigure)childF.parent(),childF);
        if (figure.hasMoreElements())
          childF.setAligner(
            getMiddleElementAligner((CompositeFigure)childF.parent(),childF,prev)
          );
        else
          childF.setAligner(
            getBottomElementAligner((CompositeFigure)childF.parent(),childF,prev)
          );
        prev = childF;
      }
    }*/
  }

  /**

   * Gets the aligner for the top element of the list.

   */

  abstract protected Aligner getTopElementAligner(CompositeFigure container,Figure component);


  /**

   * Gets the aligner for a middle element of the list.

   */

  abstract protected Aligner getMiddleElementAligner(CompositeFigure container,Figure component,Figure previous);


  /**

   * Gets the aligner for the bottom element of the list.

   */

  abstract protected Aligner getBottomElementAligner(CompositeFigure container,Figure component,Figure previous);

  /**
   * Fill the arrangement vector with Aligner objects
   * for the subfigures according to the layout strategy and current display box.
   * The default layout strategy is to leave the subfigures where they are relative
   * to the origin of the displaybox. Override in subclasses.
   */
  protected void arrange(CompositeFigure container) {
    //System.err.println("FigureHolder#arrange");
    //attachAligners(container);
    Enumeration figure = getStationaryFigures().elements();
    //Enumeration figure = fHoldees.elements();
		while (figure.hasMoreElements())
      fArrangement.addElement(
        ((Figure)figure.nextElement()).getAligner()
        //new RelativeAligner(this,(Figure)figure.nextElement())
      );
  }

  /**
   * Find the position of figure among the subfigures according to the current
   * layout strategy. Return an index to the vector of subfigures. The default
   * is to return the last+1 position.
   * => make subfigure layout a REAL strategy??
   */
  protected int findPos(Figure figure) {
    return getComponents().size(); // add last
    /* Vector figures = getFigures();
    int pos = 0;
    while (pos < figures.size() &&
           figure.displayBox().getLocation().y >
           ((Figure)figures.elementAt(pos)).displayBox().getLocation().y
           //figure.center().y >
           //((Figure)figures.elementAt(pos)).center().y
          )
      pos++;
    return pos;*/
    // moveIn() here???
  }

  /*private void printArrangement(Vector v) {
    for (int i = 0;i<v.size();i++)
      //System.err.println(i+" "+v.elementAt(i));
  }*/


   // !!!! should I always pack after accommodate and remove???


    public Dimension preferredLayoutSize(CompositeFigure container) {
    	Vector v = getArrengement(container);
      Vector holdees = getStationaryFigures();
      Insets insets = container.getInsets(); 
      //System.err.println("v: "+v);
    	if (v.isEmpty()) {
    		//System.err.println("pref size empty");
    		return emptySize();
    	}
      Point pos = ((Aligner)v.elementAt(0)).align((Figure)holdees.elementAt(0));
      Point origin = container.displayBox().getLocation();
      origin.translate(insets.left,insets.top);
      Rectangle box = ((Figure)holdees.elementAt(0)).displayBox();
      //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    	Rectangle r =   // pos.x => origin.x && pos.y => origin.y
        new Rectangle(origin,new Dimension(pos.x-origin.x,pos.y-origin.y))
          .union(new Rectangle(pos.x,pos.y,box.width,box.height));
      //System.err.println("r: "+r);
      for (int i = 1;i<holdees.size();i++) {
        pos = ((Aligner)v.elementAt(i)).align((Figure)holdees.elementAt(i));
        box = ((Figure)holdees.elementAt(i)).displayBox();
   		  r = r.union(new Rectangle(pos.x,pos.y,box.width,box.height));
      //System.err.println("r: "+r);
      }
    	//r.setSize(
    	//	r.width+insets.left+insets.right,
    	//	r.height+insets.top+insets.bottom
    	//);
    	//System.err.println(">>>> pref size "+r);
    	return new Dimension(r.width,r.height);
    }


		protected void accommodateComponents(CompositeFigure container) {
			Vector v = getArrengement(container);
      Vector holdees = getStationaryFigures();
      for (int i=0;i<v.size();i++) {
     		Figure fig = (Figure)holdees.elementAt(i);
        Point newPos = ((Aligner)v.elementAt(i)).align(fig);
        //System.err.println(">>>>>>newPos:"+newPos);
        Point oldPos = fig.displayBox().getLocation();
        //System.err.println(">>>>>>oldPos:"+oldPos);
        if (!newPos.equals(oldPos)) {
          boolean markedForMove = fig.inTransit();
            //DragTransactionManager.getInstance().isAboutToMove(fig);
          if (!markedForMove)
            fig.beginDrag(null);
    		    //DragTransactionManager.getInstance().addAboutToMove(fig);
          //System.err.println("&&&&& moving: "+fig);
    		  fig.moveBy(newPos.x - oldPos.x,newPos.y - oldPos.y);
          if (!markedForMove)
            fig.endDrag(null);
    		    //DragTransactionManager.getInstance().removeAboutToMove(fig);
        }
    		grow(container,fig.displayBox());
    		//System.err.println("pack(): figure: "+fig);
    	}
		}

    public void layoutContainer(CompositeFigure container) {
      //System.err.println("<<*>> lay out "+container);
    	/*Vector v = getArrengement(container);
      Vector holdees = getStationaryFigures();
      for (int i=0;i<v.size();i++) {
     		Figure fig = (Figure)holdees.elementAt(i);
        Point newPos = ((Aligner)v.elementAt(i)).align(fig);
        //System.err.println(">>>>>>newPos:"+newPos);
        Point oldPos = fig.displayBox().getLocation();
        //System.err.println(">>>>>>oldPos:"+oldPos);
        if (!newPos.equals(oldPos)) {
          boolean markedForMove = fig.inTransit();
            //DragTransactionManager.getInstance().isAboutToMove(fig);
          if (!markedForMove)
            fig.beginDrag(null);
    		    //DragTransactionManager.getInstance().addAboutToMove(fig);
          //System.err.println("pack: moving: "+fig);
    		  fig.moveBy(newPos.x - oldPos.x,newPos.y - oldPos.y);
          if (!markedForMove)
            fig.endDrag(null);
    		    //DragTransactionManager.getInstance().removeAboutToMove(fig);
        }
    		//System.err.println("pack(): figure: "+fig);
    	}*/
    	if (!fComponents.isEmpty()) {
    		accommodateComponents(container);
    		/*Rectangle box = 
    			new Rectangle(
    				container.displayBox().getLocation(),
    				preferredLayoutSize(container)
    			);
    		grow(container,box);*/
    	}
    	layoutSovereigns(container);
    }

  public Dimension minimumLayoutSize(CompositeFigure container) {
    return preferredLayoutSize(container);
  }
  

}