package CH.ifa.draw.contrib.xcomp;

import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import java.awt.*;
import java.util.*;
import java.io.*;


public class BagLayout implements DrawLayoutManager {
	
	/**
	 * The bag layout algorithm. Enumerates over the subfigures of the
	 * container and calls accommodate to position each child and to adjust
	 * the display box of the container so that all subfigures are
	 * enclosed.
	 */
	public void layoutContainer(CompositeFigure container) {
  	FigureEnumeration figures = container.figures();
  	while (figures.hasMoreElements()) {
  		Figure figure = figures.nextFigure();
  		//System.err.println("((( "+figure);
  		if (
  			  //!(figure instanceof Sovereign) && 
  			  !figure.isValid())
  			accommodate(container,figure);
  	}			
	}
	
	/**
   * The basic subfigure positioning algorithm. Moves the child so that the origin of its
   * display box is inside the display box of the container and then
   * makes the display box of the container to grow, if necessary, to
   * fully enclose the child. That is, the child gets placed where it
   * was dropped. Overlapping of child figures is not checked.
   */
	public void accommodate(CompositeFigure container,Figure child) {
		//System.err.println("accommodate: "+child+" ** "+container);
     		//System.err.println("====>child: "+child);
        Point pos = child.displayBox().getLocation();
        //System.err.println("====>pos:"+pos);
        Rectangle newBox = new Rectangle(pos,child.getPreferredSize());
        if (!newBox.equals(child.displayBox())) {
        	child.invalidate();
        	child.basicDisplayBox(
        		pos,
        		new Point(pos.x+newBox.width,pos.y+newBox.height)
        	);
        	child.markDirty(); // should markDirty() mark subfigures dirty too?
        	//fig.invalidate();
        }
				//moveIn(container,child);
				//grow(container,child.displayBox()); 
	}

	/**
	 * Does nothing. The default is not to translate the minimum size recatngle.
	 */
	public void overlayMinimumRectangle(CompositeFigure container,Rectangle minimum) {
		
	}

  /**
   * Move the subfigure inside of this if the subfigure extends either
   * to the left or up (or both) of the origin of this.
   */
  protected void moveIn(CompositeFigure container,Figure child) {
    Rectangle kidBox = child.displayBox();
    Rectangle myBox = container.displayBox();
    Point distance = new Point(0,0);
    Insets padding = container.getInsets();
    distance.x = kidBox.x < (myBox.x+padding.left) ? myBox.x+padding.left-kidBox.x : 0;
    distance.y = kidBox.y < (myBox.y+padding.top) ? myBox.y+padding.top-kidBox.y : 0;
    if (distance.x > 0 || distance.y > 0) {
    	//System.err.println("]]]] moving in "+child);
      boolean alreadyMoving = child.inTransit();
      if (!alreadyMoving)
        child.beginDrag(null);
      child.moveBy(distance.x,distance.y);
      if (!alreadyMoving)
        child.endDrag(null);
    }
  }

  /**
   * Grow the display box of the container so that the childBox is enclosed.
   */
	protected void grow(CompositeFigure container,Rectangle childBox) {
		childBox.setSize(
			childBox.width+container.getInsets().left+container.getInsets().right,
			childBox.height+container.getInsets().top+container.getInsets().bottom
		); // leave some space around the child
		childBox.translate(-container.getInsets().left,-container.getInsets().top);
		//System.err.println("++++ container: "+container+" "+container.displayBox());
    //System.err.println("++++ childBox: "+childBox);
		Rectangle newBox = container.displayBox().union(childBox);
		if (!newBox.equals(container.displayBox())) {
			//System.err.println(")))))) ok, newbox: "+newBox);
			//container.displayBox(
			//	new Point(newBox.x,newBox.y),
			//	new Point(newBox.x+newBox.width,newBox.y+newBox.height)
			//);
			
			//container.invalidate(); // not needed, newBox > oldBox ?
			container.basicDisplayBox(
				new Point(newBox.x,newBox.y),
				new Point(newBox.x+newBox.width,newBox.y+newBox.height)
			);
			container.invalidate();
			
		}
	}
	
	/**
	 * Does nothing.
	 */
	public void invalidateLayout() {
		//System.err.println("bag layout invalidated");
	}
	
	/**
	 * Calls invalidateLayout().
	 */
	public void invalidateLayout(Figure component) {
		invalidateLayout();
	}
	
	/**
	 * Calculates the minimum size dimensions for the specified container.
	 */ 
	public Dimension minimumLayoutSize(CompositeFigure container) {
		Rectangle r = 
			new Rectangle(
				container.displayBox().getLocation(),new Dimension(1,1)
			);
		Insets in = container.getInsets();
		FigureEnumeration figures = container.figures();
		while (figures.hasMoreElements()) {
			Figure child = figures.nextFigure();
			if (!(child instanceof Sovereign))
				r.add(child.displayBox());
		}
		r.grow(in.right/2,in.bottom/2);
		return r.getSize();
	}

	/**
	 * Calculates the minimum size display box for the specified container
	 * at its current location.
	 */ 
	public Rectangle minimumLayoutRectangle(CompositeFigure container) {
		return new Rectangle(
			container.displayBox().getLocation(),minimumLayoutSize(container)
		);
	}

	/**
	 * Calculates the preferred size dimensions for the specified container.
	 */ 
	public Dimension preferredLayoutSize(CompositeFigure container) {
		return minimumLayoutSize(container);
	}

	/**
	 * Calculates the minimum size display box for the specified container
	 * at its current location.
	 */ 
	public Rectangle preferredLayoutRectangle(CompositeFigure container) {
		return new Rectangle(
			container.displayBox().getLocation(),preferredLayoutSize(container)
		);
	}

	/**
	 * Does nothing. Bag layout does not need to store the components.
	 */
	public void addLayoutComponent(Figure component) {}	
	/**
	 * Does nothing. Bag layout does not need to store the components.
	 */
	public void removeLayoutComponent(Figure component) {}

	/**
	 * Does nothing. Bag layout does not need to store the components.
	 */
	public void replaceLayoutComponent(Figure component, Figure replacement) {}
	
}