

package CH.ifa.draw.figures;

import CH.ifa.draw.framework.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.contrib.xcomp.*;
import CH.ifa.draw.util.Geom;
import java.awt.*;
import java.util.*;
import java.io.*;

/**
 * A LabelledConnection is a LineConnection that maintains a list of
 * LabelledSegments in addition to the list of points. A segment is a holder
 * for Labels objects (CompartmentPanel figures) that hold the labels.
 * A LabelledConnection forwards
 * the selection and the drop target finding events to the Labels.<p>
 * The segment list is kept consistent with the point list. When segments are
 * removed, the Labels of the removed segment are merged with a nearby segment.
 * The creation of the first segment is delayed until the connection is 'made'
 * (handleConnect()) for the first time. This is to ensure that the connection
 * has been properly set up (added to the drawing etc.) before creating the
 * segment object and its label holder figures.
 */
public class LabelledConnection extends LineConnection {
  /** The segments of this. */
  private Vector fSegments;

  /** Deprecated. */
  private Vector fLabels;

  public LabelledConnection() {
    super();
    fSegments = new Vector();
    fLabels = new Vector();
    setStartDecoration(null);
  }

 /**
  * Set the parent of this. Inform listeners about the change. Also invoke
  * ancestorChanged to inform a change in ancestry.
  */
	public void parent(Figure aFigure) {
    super.parent(aFigure);
    Enumeration segment = fSegments.elements();
    while (segment.hasMoreElements()) {
      //((Figure)segment.nextElement()).setParent(aFigure); ????
      LabelledSegment s = (LabelledSegment)segment.nextElement();
      for (int i = s.first();i<=s.last();i++)
        ParentChangeEvent.changeParent(s.getLabelAt(i),aFigure);
    }
	}

  /**
   * Just set the parent without informing listeners about the change.
   * This is needed for de-serialization.
   */
	public void setParent(Figure aFigure) {
		super.setParent(aFigure);
    Enumeration segment = fSegments.elements();
    while (segment.hasMoreElements()){
      //((Figure)segment.nextElement()).setParent(aFigure); ????
      LabelledSegment s = (LabelledSegment)segment.nextElement();
      for (int i = s.first();i<=s.last();i++)
        ParentChangeEvent.changeParent(s.getLabelAt(i),aFigure);
    }
	}

  /** Deprecated. No need to use this method. */
  public Figure addConnected(Figure figure) {
    if (!fLabels.contains(figure))
      fLabels.addElement(figure);
    return figure;
  }

  /** Deprecated. No need to use this method. */
  public Figure removeConnected(Figure figure) {
    if (fLabels.contains(figure))
      fLabels.removeElement(figure);
    return figure;
  }

  /** Deprecated. No need to use this method. */
  public Vector getLabels() {
    return (Vector)fLabels.clone();
  }

  /** Returns the segments of this. */
  public Vector getSegments() {
    return (Vector)fSegments.clone();
  }

    /**
     * Returns the locator used to locate connected text. Deprecated?
     */
    public Locator connectedTextLocator(Figure text) {
      if (text instanceof TextFigure && ((TextFigure)text).isConnected() && text.inTransit()) {
        int segment = SegmentLocator.findSegment(this,text);
        if (segment == -1)
          return null;
        else
          return new SlotLocator(this,text,segment);
      }
      return new SlotLocator(this,text);
    }

    /**
     * Adds a node to the list of points and the corresponding segment.
     */
    public void addPoint(int x, int y) {
      super.addPoint(x,y);
      fSegments.addElement(createSegment(pointCount()-1));
    }

    /**
     * Creates the first segment, if segments do not yet exists.
     */
    protected void handleConnect(Figure start, Figure end) {
    	//f (fSegments == null) fSegments = new Vector();
    	if (fSegments != null) {
      	super.handleConnect(start,end);
      	//if (fSegments != null && fSegments.isEmpty()) {
      	if (fSegments.isEmpty()) {
        //System.err.println("fSegments.isEmpty()");
        	fSegments.addElement(createSegment(0));
      	}
      }
    }

    /**
     * Check first the labels of the segments and then do the default
     * check.
     */
    public Figure findDropTarget(int x, int y, Figure dragTarget) {
      Enumeration segment = fSegments.elements();
      while (segment.hasMoreElements()) {
      //((Figure)segment.nextElement()).setParent(aFigure); ????
        LabelledSegment s = (LabelledSegment)segment.nextElement();
        for (int i = s.first();i<=s.last();i++) {
          Figure f = s.getLabelAt(i).findDropTarget(x,y,dragTarget);
          if (f != null)
            return f;
        }
      }
      return super.findDropTarget(x,y,dragTarget);
    }

    /**
     * Check first the labels of the segments and then do the default
     * find.
     */
    public Figure findSelectable(int x, int y) {
      Enumeration segment = fSegments.elements();
      while (segment.hasMoreElements()) {
      //((Figure)segment.nextElement()).setParent(aFigure); ????
        LabelledSegment s = (LabelledSegment)segment.nextElement();
        for (int i = s.first();i<=s.last();i++) {
          Figure f = s.getLabelAt(i).findSelectable(x,y);
          if (f != null)
            return f;
        }
      }
      return super.findSelectable(x,y);
    }

    /**
     * Insert a node at the given point and create the corresponding segment.
     */
    public void insertPointAt(Point p, int i) {
      if (fSegments.isEmpty()) {
        //System.err.println("fSegments.isEmpty()");
        fSegments.addElement(createSegment(0));
      }
      //System.err.println("i: "+i);
      fPoints.insertElementAt(p, i); //from PolyLineFigure
      fSegments.insertElementAt(createSegment(i),i);
      for (int j = i+1; j < pointCount()-1;j++)
        ((LabelledSegment)fSegments.elementAt(j)).setSegment(j);

      changed(); //from PolyLineFigure
      layoutConnection(); //from LineConnection
    }
    /**
     * Remove a point and the corresponding segment. Merge the labels
     * of the removed segment with the next segment.
     */
    public void removePointAt(int i) {
      //System.err.println("removePointAt: "+i);
      //System.err.println("removeSegmentAt: "+(i-1));
      LabelledSegment removed = (LabelledSegment)fSegments.elementAt(i-1);
      fSegments.removeElementAt(i-1);
      for (int j = i-1; j < pointCount()-2;j++) {
        //System.err.println("j: "+j);
        ((LabelledSegment)fSegments.elementAt(j)).setSegment(j);
      }
      ((LabelledSegment)fSegments.elementAt(i-1)).merge(removed);
        // merge the labels of the removed segment with the labels of the next
        // segment and decrement the position of the segements after i. 
      super.removePointAt(i);
    }


    /*protected void handleBroughtToFrontEvent(BroughtToFrontEvent e) {

      if (e.getSource() != this && ) {
        //System.err.println("connectee brought to front");
        parent().bringToFront(this);
      }
    }*/

    /**
     * A hook for creating the a segment object. The default is to create
     * a Segment which a local class.
     */
    protected LabelledSegment createSegment(int segmentIndex) {
      return new Segment(this,segmentIndex);
    }
    
  	private void readObject(ObjectInputStream s)
        throws ClassNotFoundException, IOException {

  		s.defaultReadObject();

    	Enumeration segment = fSegments.elements();
    	while (segment.hasMoreElements()){
      //((Figure)segment.nextElement()).setParent(aFigure); ????
      	LabelledSegment se = (LabelledSegment)segment.nextElement();
     		//System.err.println("##Segment "+se);
     		for (int i = se.first();i<=se.last();i++) {
     			//System.err.println("##Label "+se.getLabelAt(i));
        	addFigureChangeListener(se.getLabelAt(i));
    		}
   		}
    }
}

/**
 * The default segment implementation. A segment has six Labels at the
 * beginning, middle, and end of the segment on both sides. A segment knows
 * its owner connection and its relative position in the connection.
 */
class Segment implements LabelledSegment,Serializable {

  /* Constants for the positions of the labels. */
  static final int StartLeft = 0;
  static final int StartRight = 1;
  static final int CenterLeft = 2;
  static final int CenterRight = 3;
  static final int EndLeft = 4;
  static final int EndRight = 5;
  /** The relative position of this in the owner. */
  int fSegment;
  /** The owner of this. */
  LabelledConnection fOwner;
  /** The Labels of this. */
  Vector fLabels;

  public int first() {
    return StartLeft;
  }

  public int last() {
    return EndRight;
  }

  public Segment(LabelledConnection owner, int segment) {
    fOwner = owner;
    fSegment = segment;
    fLabels = new Vector(6);
    createLabelSlots();
  }

  public void setSegment(int segment) {
    Enumeration label = fLabels.elements();
    while (label.hasMoreElements()) {
      LabelPanel l = (LabelPanel)label.nextElement();
      if (l != null)
        ((SlotLocator)l.getLocator()).setSegment(segment);
    }
    fSegment = segment;
  }

  public LabelPanel getLabelAt(int index) {
    //0 <= index <= 5
    return (LabelPanel)fLabels.elementAt(index);
  }

  /**
   * Changes the parent of label to the parent of the owner and connects
   * the label to the parent. Calls mapToSlot() to get the locator that
   * computes the position of the label relative to the owner.
   */
  public void setLabelAt(LabelPanel label,int index) {
    //0 <= index <= 5
    fLabels.setElementAt(label,index);
    ParentChangeEvent.changeParent(label,fOwner.parent());
    label.connect(
      fOwner,
      mapToSlot(index,label)
    );
  }

  /**
   * Creates a Labels object at the relative position given by index.
   */
  public void createLabelAt(int index) {
    //0 <= index <= 5
    LabelPanel label = new LabelPanel();
    label.displayBox(new Point(fOwner.displayBox().getLocation()),new Point(10,10));
    setLabelAt(label,index);
  }

  /**
   * Returns a SlotLocator corresponding to the relative location (begin-left,
   * begin-right, middle-left, middle-right, end-left, end-right) as determined
   * by the parameter index.
   */
  protected SlotLocator mapToSlot(int index, LabelPanel label) {
    switch (index) {
      case StartLeft :
      {
        return new SlotLocator(
            fOwner,label,fSegment,
            SlotLocator.getBeginSlotLocator(fOwner,label,true)
        );
        //break;
      }
      case StartRight :
      {
        return new SlotLocator(
            fOwner,label,fSegment,
            SlotLocator.getBeginSlotLocator(fOwner,label,false)
          );
        //break;
      }
      case CenterLeft :
      {
        return new SlotLocator(
            fOwner,label,fSegment,
            SlotLocator.getCenterSlotLocator(fOwner,label,true)
          );
        //break;
      }
      case CenterRight :
      {
        return new SlotLocator(
            fOwner,label,fSegment,
            SlotLocator.getCenterSlotLocator(fOwner,label,false)
          );
        //break;
      }
      case EndLeft :
      {
        return new SlotLocator(
            fOwner,label,fSegment,
            SlotLocator.getEndSlotLocator(fOwner,label,true)
          );
        //break;
      }
      case EndRight :
      {
        return new SlotLocator(
            fOwner,label,fSegment,
            SlotLocator.getEndSlotLocator(fOwner,label,false)
          );
        //break;
      }
      default :
        throw new ArrayIndexOutOfBoundsException("Invalid slot index "+index);
    }
  }

  protected void createLabelSlots() {
    for (int i = first(); i <= last(); i++) {
      fLabels.addElement(null);
      createLabelAt(i);
    }
  }

  /**
   * Merge the contents of the labels of the segment other with the labels
   * of this. The actual merging of the contents (the actual figures used
   * as the labels) of two Labels-holders is delegated to Labels.
   */
  public void merge(LabelledSegment other) {
    other.setSegment(fSegment);   // still needed ??
    for (int i=first();i<=last();i++) {
    	LabelPanel l = other.getLabelAt(i);
      getLabelAt(i).merge(l);
      l.disconnect();
      ((CompositeFigure)l.parent()).remove(l);
    }
  }
  
}