/*
 * @(#)AbstractFigure.java 5.1
 *
 */

package CH.ifa.draw.standard;

import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

/**
 * AbstractFigure provides default implementations for
 * the Figure interface.
 *
 * <hr>
 * <b>Design Patterns</b><P>
 * <img src="images/red-ball-small.gif" width=6 height=6 alt=" o ">
 * <b><a href=../pattlets/sld036.htm>Template Method</a></b><br>
 * Template Methods implement default and invariant behavior for
 * figure subclasses.
 * <hr>
 *
 * @see Figure
 * @see Handle
 */

public abstract class AbstractFigure implements Figure {

    /**
     * The listeners for a figure's changes.
     * @see #invalidate
     * @see #changed
     * @see #willChange
     */
    protected transient FigureChangeListener fListener;

    /*
     * Serialization support.
     */
    private static final long serialVersionUID = -10857585979273442L;
    private int abstractFigureSerializedDataVersion = 1;
    
    /**
     * The parent composite of this figure.
     */
     
    private transient Figure fParent;

    /**
     * If this is true, the figure may draw itself in a special way
     * to emphasize itself.
     */
    protected boolean fIsHighlighted = false;

    /**
     * If this is true, the figure is currently being dragged.
     */
    protected boolean fInTransit = false;

    protected boolean fHasMoved = false;

    /**
     * If this is not null, the figure is currently being manipulated
     * (by a handle).
     */
    protected transient Handle fManipulator = null;

    /**
     * The aligner object that a figure holder (compartment) associates
     * with a figure it holds.
     */
    protected Aligner fAligner = null;

    protected AbstractFigure() { }

    /**
     * Set fIsHighlighted to the new value if it is different from
     * the current value. Invalidates this if the value is actually
     * changed.
     */
    public void setHighlighted(boolean isHighlighted) {
    	if (isHighlighted != fIsHighlighted) {
        //System.err.println("setHighlight = "+isHighlighted+" "+this);
    		fIsHighlighted = isHighlighted;
        invalidate();
    		//changed();
    	}
    }

    /**
     * Draw this in a special way. The default is to do nothing.
     */
    protected void highlight(Graphics g) {
    }

 /**
  * Set the parent of this. Inform listeners about the change. Also invoke
  * ancestorChanged to inform a change in ancestry.
  */
	public void parent(Figure aFigure) {
		fParent = aFigure;
    if (fListener != null)
      fListener.figureChanged(new ParentChangeEvent(this));
    ancestorChanged();
	}

  /**
   * Just set the parent without iforming listeners about the change.
   * This is needed for de-serialization.
   */
	public void setParent(Figure aFigure) {
		fParent = aFigure;
	}

  /**
   * Answer the parent of this.
   */
	public Figure parent() {
		return fParent;
	}

  /**
   * Invalidate the child.
   */
  public void accommodate(Figure child) {
    child.invalidate();
  }

    /**
     * Returns true if figure can be dropped onto this. By default figures
     * do not accept drops.
     */
  public boolean accept(Figure dropped) {
    return false;
  }

    /**
     * Drop a figure onto this. By default do nothing.
     *
     * @param the figure to drop
     * @return the dropped Figure
     */
  public Figure drop(Figure dropped) {
    return dropped;
  }


  /**
   * Inform listeners about a change in the ancestry of this.
   */
	public void ancestorChanged() {
    if (fListener != null)
      fListener.figureChanged(new AncestorChangedEvent(this));
	}

	public void unconnectedBy(ConnectionFigure connection) { }

    /**
     * Moves the figure by the given offset.
     */
    public void moveBy(int dx, int dy) {
        willChange();
        basicMoveBy(dx, dy);
        changed();
        moved(dx,dy);
    }

    /**
     * Moves the figure. This is the
     * method that subclassers override. Clients usually
     * call displayBox.
     * @see moveBy
     */
    public abstract void basicMoveBy(int dx, int dy);

    /**
     * Changes the display box of a figure. Clients usually
     * call this method. It changes the display box
     * and announces the corresponding change.
     * @param origin the new origin
     * @param corner the new corner
     * @see displayBox
     */
    public void displayBox(Point origin, Point corner) {
        if (parent() != null)
          if (!((CompositeFigure)parent()).acceptSubFigureChange(origin,corner))
            return;
        willChange();
        basicDisplayBox(origin, corner);
        changed();
    }

    /**
     * Sets the display box of a figure. This is the
     * method that subclassers override. Clients usually
     * call displayBox.
     * @see displayBox
     */
    public abstract void basicDisplayBox(Point origin, Point corner);

    /**
     * Gets the display box of a figure.
     */
    public abstract Rectangle displayBox();

    /**
     * Returns the handles of a Figure that can be used
     * to manipulate some of its attributes.
     * @return a Vector of handles
     * @see Handle
     */
    public abstract Vector handles();

    /**
     * Returns an Enumeration of the figures contained in this figure.
     * @see CompositeFigure
     */
    public FigureEnumeration figures() {
        Vector figures = new Vector(1);
        figures.addElement(this);
        return new FigureEnumerator(figures);
    }

    public FigureEnumeration figuresReverse() {
        return figures();
    }

    /**
     * Gets the size of the figure. A convenience method.
     */
    public Dimension size() {
        return new Dimension(displayBox().width, displayBox().height);
    }

    /**
     * Checks if the figure is empty. The default implementation returns
     * true if the width or height of its display box is < 3
     * @see Figure#isEmpty
     */
    public boolean isEmpty() {
        return (size().width < 3) || (size().height < 3);
    }

    /**
     * Returns the figure that contains the given point.
     * In contrast to containsPoint it returns its
     * innermost figure that contains the point.
     *
     * @see #containsPoint
     */
    public Figure findFigureInside(int x, int y) {
        if (containsPoint(x, y))
            return this;
        return null;
    }

    /**
     * Returns the selectable figure that contains the given point.
     * [By default this is 'flat' (non-nested component)].
     */
    public Figure findSelectable(int x, int y) {
      //System.err.println("findSelectable "+this);
        if (containsPoint(x, y))
            return this;
        return null;
    }

    /**
     * Returns null. By default, figures are not control-down selectable.
     */
    public Figure findCtrlSelectable(int x, int y) {
        return null;
    }

   /**
     * Returns the connection figure that contains the given point.
     * The default is to return null. Connection figures should
     * override this.
     */
    public Figure findConnection(int x, int y) {
      return null;
    }

    /**
     * Returns the connectable figure that contains the given point.
     * The default implementation is to check whether this contains the
     * point and can be connected.
     */
    public Figure findConnectable(int x, int y,Figure connection) {
      if (containsPoint(x,y) && canConnect() 
      	 // && !includes(connection)
      	  )
        return this;
      return null;
    }

    /**
     * Default implementation: return this if this is not the dragTarget,
     * this contains point (x,y) and this accept()s the target. Otherwise,
     * returns null;
     */
    public Figure findDropTarget(int x, int y, Figure dragTarget) {
      if (dragTarget != this && containsPoint(x,y) && accept(dragTarget))
        return this;
      return null;
    }

    /**
     * Checks if a point is inside the figure.
     */
    public boolean containsPoint(int x, int y) {
        return displayBox().contains(x, y);
    }

    /**
     * Changes the display box of a figure. This is a
     * convenience method. Implementors should only
     * have to override basicDisplayBox
     * @see displayBox
     */
    public void displayBox(Rectangle r) {
        displayBox(new Point(r.x, r.y), new Point(r.x+r.width, r.y+r.height));
    }

    /**
     * Checks whether the given figure is contained in this figure.
     */
    public boolean includes(Figure figure) {
        return figure == this;
    }

    /**
     * Decomposes a figure into its parts. It returns a Vector
     * that contains itself.
     * @return an Enumeration for a Vector with itself as the
     * only element.
     */
    public FigureEnumeration decompose() {
        Vector figures = new Vector(1);
        figures.addElement(this);
        return new FigureEnumerator(figures);
    }

    /**
     * Sets the Figure's container and registers the container
     * as a figure change listener. A figure's container can be
     * any kind of FigureChangeListener. A figure is not restricted
     * to have a single container.
     */
    public void addToContainer(FigureChangeListener c) {
        addFigureChangeListener(c);
        invalidate();
    }

    /**
     * Removes a figure from the given container and unregisters
     * it as a change listener.
     */
    public void removeFromContainer(FigureChangeListener c) {
        invalidate();
        removeFigureChangeListener(c);
        changed();
    }

    /**
     * Adds a listener for this figure.
     */
    public void addFigureChangeListener(FigureChangeListener l) {
        fListener = FigureChangeEventMulticaster.add(fListener, l);
    }

    /**
     * Removes a listener for this figure.
     */
    public void removeFigureChangeListener(FigureChangeListener l) {
        fListener = FigureChangeEventMulticaster.remove(fListener, l);
    }

    /**
     * Gets the figure's listners.
     */
    public FigureChangeListener listener() {
        return fListener;
    }

    /**
     * A figure is released from the drawing. You never call this
     * method directly. Release notifies its listeners.
     * @see Figure#release
     */
    public void release() {
        if (fListener != null)
            fListener.figureRemoved(new FigureChangeEvent(this));
    }

    /**
     * Invalidates the figure. This method informs the listeners
     * that the figure's current display box is invalid and should be
     * refreshed.
     */
    public void invalidate() {
        if (fListener != null) {
            Rectangle r = displayBox();
            r.grow(Handle.HANDLESIZE, Handle.HANDLESIZE);
            fListener.figureInvalidated(new FigureChangeEvent(this, r));
        }
    }

    /**
     * Informes that a figure is about to change something that
     * affects the contents of its display box.
     *
     * @see Figure#willChange
     */
    public void willChange() {
        invalidate();
    }

    /**
     * Informs that a figure changed the area of its display box.
     *
     * @see FigureChangeEvent
     * @see Figure#changed
     */
    public void changed() {
        invalidate();
        if (fListener != null) {
        	//	System.err.println("announcing change of "+this.toString());
            fListener.figureChanged(new FigureChangeEvent(this));
        }

    }

    /**
     * Informs that a figure has moved.
     *
     * @see FigureMovedEvent
     * @see Figure#moved
     */
    public void moved(int dx,int dy) {
        if (fListener != null)
            fListener.figureChanged(new FigureMovedEvent(this,dx,dy));
    }

  /**
   *  Returns true. By default, figures do not need to be laid out.
   */
  public boolean isValid() {
  	return true;
  }
  
  /**
   * Announce a change
   */
  public void markDirty() {
  	changed();
  }
    
  /**
   * Calls markDirty().
   */
  public void markDirty(Figure figure) {
  	markDirty();
  }

  /**
   * Does nothing. By default, figures have always a valid layout.
   */
  public void markValid() {
  }
  
  /**
   * By default, does nothing. Containers (composites) that maintain
   * a lay out of subfigures should implement this.
   */
  public void validate() {
  //	System.err.println("||| validating "+this);
  }
  
  /**
   * By default, does nothing. Containers (composites) that maintain
   * a lay out of subfigures should implement this.
   */
  public void doLayout() {	
  //	System.err.println("!!! lay out "+this);
  }
  
    /**
     * Bring the child (sub-)figure to the front of the display list
     * (i.e. to be drawed last). By default figures do not have subfigures
     * and the request is ignored.
     */
    public void bringToFront(Figure child) {
      //System.err.println(">>>AbstractF#bringToFront "+child);
    }
    
    /**
     * Informs that a figure has been brought to front.
     *
     * @see BroughtToFrontEvent
     * @see Figure#broughtToFront
     */
     public void broughtToFront() {
        if (fListener != null)
            fListener.figureChanged(new BroughtToFrontEvent(this));
    }

    /**
     * Gets the center of a figure. A convenience
     * method that is rarely overridden.
     */
    public Point center() {
        return Geom.center(displayBox());
    }

    /**
     * Checks if this figure can be connected. By default
     * AbstractFigures can be connected.
     */
    public boolean canConnect() {
        return true;
    }

    /**
     * Returns the connection inset. The connection inset
     * defines the area where the display box of a
     * figure can't be connected. By default the entire
     * display box can be connected.
     *
     */
    public Insets connectionInsets() {
        return new Insets(0, 0, 0, 0);
    }

    /**
     * Returns the Figures connector for the specified location.
     * By default a ChopBoxConnector is returned.
     * @see ChopBoxConnector
     */
    public Connector connectorAt(int x, int y) {
        return new ChopBoxConnector(this);
    }

    /**
     * Sets whether the connectors should be visible.
     * By default they are not visible and
     */
    public void connectorVisibility(boolean isVisible) {
    }

    /**
     * Returns the locator used to located connected text.
     */
    public Locator connectedTextLocator(Figure text) {
        return RelativeLocator.center();
    }

    /**
     * Returns the named attribute or null if a
     * a figure doesn't have an attribute.
     * By default
     * figures don't have any attributes getAttribute
     * returns null.
     */
    public Object getAttribute(String name) {
        return null;
    }

    /**
     * Sets the named attribute to the new value. By default
     * figures don't have any attributes and the request is ignored.
     */
    public void setAttribute(String name, Object value) {
    }

    /**
     * Clones a figure. Creates a clone by using the storable
     * mechanism to flatten the Figure to stream followed by
     * resurrecting it from the same stream.
     *
     * @see Figure#clone
     */
    public Object clone() {
        Object clone = null;
        ByteArrayOutputStream output = new ByteArrayOutputStream(200);
        try {
            ObjectOutput writer = new ObjectOutputStream(output);
            writer.writeObject(this);
            writer.close();
        } catch (IOException e) {
            System.out.println("Class not found: " + e);
        }

        InputStream input = new ByteArrayInputStream(output.toByteArray());
        try {
            ObjectInput reader = new ObjectInputStream(input);
            clone = (Object) reader.readObject();
        } catch (IOException e) {
            System.out.println(e.toString());
        }
        catch (ClassNotFoundException e) {
            System.out.println("Class not found: " + e);
        }
        return clone;
    }

    /**
     * Stores the Figure to a StorableOutput.
     */
    public void write(StorableOutput dw) {
    }

    /**
     * Reads the Figure from a StorableInput.
     */
    public void read(StorableInput dr) throws IOException {
    }

    public boolean isFlat() {
    	return true;
    }
  /**
   * Tell this that it is about to be dragged.
   */
  public void beginDrag(MouseEvent e) {
    fInTransit = true;
    fHasMoved = false;
    if (parent() != null)
      parent().beginDrag(this,e);
  }

  /**
   * Tell this that a child is about to be dragged.
   */
  public void beginDrag(Figure child,MouseEvent e) {
  }

  /**
   * Tell this that a child has been dragged.
   */
  public void endDrag(Figure child,MouseEvent e) {
  }

  /**
   * Tell this that it has been dragged.
   */
  public void dragged(MouseEvent e) {
    fHasMoved = true;
  }

  public boolean hasMoved() {
    return fHasMoved;
  }

  /**
   * The drag transaction is over.
   */
  public void endDrag(MouseEvent e) {
    fInTransit = false;
    fHasMoved = false;
    if (parent() != null)
      parent().endDrag(this,e);
  }

  /**
   * Answer true if this is the target of a drag transaction at the
   * moment (being dragged).
   */
  public boolean inTransit() {
    return fInTransit;
  }

  /**
   * Returns the desired size of this. The default implementation just
   * returns the size of the current display box.
   */
  public Dimension getPreferredSize() {
    return new Dimension(displayBox().width,displayBox().height);
  }

  /**
   * Returns the minimum size of this. The default implementation 
   * returns a dimension of zero width and height.
   */
  public Dimension getMinimumSize() {
    return new Dimension(0,0);
  }

  /**
   * Use this method to set the manipulator of this figure. Typically,
   * this is called upon invokeStart() by the handle that initiates the
   * manipulation. The handle should call this again with null as the
   * argument upon invokeEnd().
   */
  public void setManipulator(Handle handle) {
    fManipulator = handle;
  }

  /**
   * Use this method to get the current manipulator of this figure.
   */
  public Handle getManipulator() {
    return fManipulator;
  }

  /**
   * Associates an aligner with this figure.
   */
  public void setAligner(Aligner aligner) {
    fAligner = aligner;
  }

  /**
   * Gets the aligner associated with this figure.
   */
  public CH.ifa.draw.framework.Aligner getAligner() {
    return fAligner;
  }
  
  /**
   * Does nothing;
   */
  public void created() {
  }

}