/*
 * @(#)CompositeFigure.java 5.1
 *
 */

package CH.ifa.draw.standard;

import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.contrib.xcomp.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

/**
 * A Figure that is composed of several figures. A CompositeFigure
 * doesn't define any layout behavior. It is up to subclassers to
 * arrange the contained figures.
 * <hr>
 * <b>Design Patterns</b><P>
 * <img src="images/red-ball-small.gif" width=6 height=6 alt=" o ">
 * <b><a href=../pattlets/sld012.htm>Composite</a></b><br>
 * CompositeFigure enables to treat a composition of figures like
 * a single figure.<br>
 * @see Figure
 */

public abstract class CompositeFigure
                extends AbstractFigure
                implements FigureChangeListener {

    /**
     * The figures that this figure is composed of
     * @see #add
     * @see #remove
     */
    protected Vector fFigures;
    
    /**
     * Validation status flag. The flag is set to false, when the
     * composite needs to be validated. 
     */
    protected boolean fIsValid = true;
    
    /**
     * The layout manager of this composite.
     */
    protected DrawLayoutManager fLayoutManager;

    /*
     * Serialization support.
     */
    private static final long serialVersionUID = 7408153435700021866L;
    private int compositeFigureSerializedDataVersion = 1;

    protected CompositeFigure() {
        fFigures = new Vector();
    }
    
    // This should be made abstract??
    protected DrawLayoutManager getDefaultLayoutManager() {
   		return new CH.ifa.draw.contrib.xcomp.BagLayout();
    } 
    
    public DrawLayoutManager getLayout() {
    	if (fLayoutManager == null) 
    		fLayoutManager = getDefaultLayoutManager();	
    	//System.err.println("== == layout is "+fLayoutManager);
    	return fLayoutManager;
    }
    
    public void setLayout(DrawLayoutManager manager) {
    	fLayoutManager = manager;
    }	

    /**
     * Retrun true if the new display box as defined by origin and
     * corner of a subfigure is acceptable. The default is to accept
     * all changes. Override this to specify more restrictive behavior
     * in concrete subclasses.
     */
    public boolean acceptSubFigureChange(Point origin, Point corner) {
      return true;
    }

    /**
     * Adds a figure to the list of figures. Initializes the
     * the figure's container.
     */
    public Figure add(Figure figure) {
        if (!fFigures.contains(figure)) {
            fFigures.addElement(figure);
            figure.addToContainer(this);
            if (figure.parent() != this)
            	figure.parent(this);
            if (!(figure instanceof Sovereign))
            getLayout().addLayoutComponent(figure);
            markDirty(figure);
        }
        return figure;
    }

    /**
     * Adds a vector of figures.
     * @see #add
     */
    public void addAll(Vector newFigures) {
        Enumeration k = newFigures.elements();
        while (k.hasMoreElements())
            add((Figure) k.nextElement());
    }

    /**
     * Removes a figure from the composite. But does not release it (A-PT).
     * @see #removeAll
     */
    public Figure remove(Figure figure) {
        if (fFigures.contains(figure)) {
            getLayout().removeLayoutComponent(figure);
            figure.removeFromContainer(this);
            fFigures.removeElement(figure);
          //  System.err.println("@"+this+" removed: "+figure);
            markDirty();
        }
        return figure;
    }

    /**
     * Removes a child figure from the composite (A-PT).
     */
    public Figure removeChild(Figure figure) {
        return remove(figure);
    }

    /**
     * Removes a vector of figures.
     * @see #remove
     */
    public void removeAll(Vector figures) {
        Enumeration k = figures.elements();
        while (k.hasMoreElements())
            remove((Figure) k.nextElement());
    }

    /**
     * Removes all children.
     * @see #remove
     */
    public void removeAll() {
        FigureEnumeration k = figures();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            getLayout().removeLayoutComponent(figure);
           figure.removeFromContainer(this);
        }
        fFigures.removeAllElements();
        markDirty();
    }

    /**
     * Removes a figure from the figure list, but
     * doesn't release it. Use this method to temporarily
     * manipulate a figure outside of the drawing.
     */
    public synchronized Figure orphan(Figure figure) {
        fFigures.removeElement(figure);
        getLayout().removeLayoutComponent(figure);
        return figure;
    }

    /**
     * Removes a vector of figures from the figure's list
     * without releasing the figures.
     * @see orphan
     */
    public void orphanAll(Vector newFigures) {
        Enumeration k = newFigures.elements();
        while (k.hasMoreElements())
            orphan((Figure) k.nextElement());
    }

    /**
     * Replaces a figure in the drawing without
     * removing it from the drawing.
     */
    public synchronized void replace(Figure figure, Figure replacement) {
        int index = fFigures.indexOf(figure);
        if (index != -1) {
            replacement.addToContainer(this);   // will invalidate figure
            figure.changed();
            fFigures.setElementAt(replacement, index);
            getLayout().removeLayoutComponent(figure);
            getLayout().addLayoutComponent(replacement);
            markDirty();
        }
    }

    /**
     * Informs that a figure has been brought to front. Forwards the event
     * to subfigures in case they need to inform there listeners (such as
     * connections spanning figure hierarchies).
     *
     * @see AbstractFigure#broughtToFront
     */
    public void broughtToFront() {
        super.broughtToFront();
        FigureEnumeration child = figures();
        while (child.hasMoreElements())
        	child.nextFigure().broughtToFront();
    }

    /**
     * Informs that the ancestry of this has changed. Forwards the event
     * to subfigures in case they need to inform there listeners (such as
     * connections spanning figure hierarchies).
     *
     * @see AbstractFigure#ancestorChanged
     */
    public void ancestorChanged() {
        super.ancestorChanged();
        FigureEnumeration child = figures();
        while (child.hasMoreElements())
        	child.nextFigure().ancestorChanged();
    }

    /**
     * Sends a figure to the back of the drawing.
     */
    public synchronized void sendToBack(Figure figure) {
        if (fFigures.contains(figure)) {
            fFigures.removeElement(figure);
            fFigures.insertElementAt(figure,0);
            figure.changed();
        }
    }

    /**
     * Brings a child figure to the front (moves last in the display list).
     * Asks the parent to bring this to front if the child is not a connection
     * (would cause a loop).
     */
    public synchronized void bringToFront(Figure figure) {
        if (fFigures.contains(figure)) {
            //System.err.println(">>>CompositeF#bringToFront "+figure);
            fFigures.removeElement(figure);
            fFigures.addElement(figure);
            //figure.changed();
            if (!(figure instanceof Sovereign)) {
              if (parent() != null)
                parent().bringToFront(this);
              else
                changed(); // just to make sure
            }
        }
    }

    /**
     * Draws all the contained figures
     * @see Figure#draw
     */
    public void draw(Graphics g) {
        FigureEnumeration k = figures();
        while (k.hasMoreElements())
            k.nextFigure().draw(g);
    }


		/**
		 * Gets the insets of this composite. Default is to return
		 * zero insets. 
		 */
	public Insets getInsets() {
		return new Insets(0,0,0,0)	;
	}
		 
    /**
     * Gets a figure at the given index.
     */
    public Figure figureAt(int i) {
        return (Figure)fFigures.elementAt(i);
    }

    /**
     * Returns an Enumeration for accessing the contained figures.
     * The figures are returned in the drawing order.
     */
    public final FigureEnumeration figures() {
        return new FigureEnumerator(fFigures);
    }

    /**
     * Gets number of child figures.
     */
    public int figureCount() {
        return fFigures.size();
    }

    /**
     * Returns an Enumeration for accessing the contained figures
     * in the reverse drawing order.
     */
    public final FigureEnumeration figuresReverse() {
        return new ReverseFigureEnumerator(fFigures);
    }

    /**
     * Finds a top level Figure. Use this call for hit detection that
     * should not descend into the figure's children.
     */
    public Figure findFigure(int x, int y) {
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            if (figure.containsPoint(x, y))
                return figure;
        }
        return null;
    }

    /**
     * Finds a top level Figure that intersects the given rectangle.
     */
    public Figure findFigure(Rectangle r) {
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            Rectangle fr = figure.displayBox();
            if (r.intersects(fr))
                return figure;
        }
        return null;
    }

    /**
     * Finds a top level Figure, but supresses the passed
     * in figure. Use this method to ignore a figure
     * that is temporarily inserted into the drawing.
     * @param x the x coordinate
     * @param y the y coordinate
     * @param without the figure to be ignored during
     * the find.
     */
    public Figure findFigureWithout(int x, int y, Figure without) {
        if (without == null)
            return findFigure(x, y);
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            if (figure.containsPoint(x, y) && !figure.includes(without))
                return figure;
        }
        return null;
    }

    /**
     * Finds a top level Figure that intersects the given rectangle.
     * It supresses the passed
     * in figure. Use this method to ignore a figure
     * that is temporarily inserted into the drawing.
     */
    public Figure findFigure(Rectangle r, Figure without) {
        if (without == null)
            return findFigure(r);
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            Rectangle fr = figure.displayBox();
            if (r.intersects(fr) && !figure.includes(without))
                return figure;
        }
        return null;
    }

    /**
     * Finds a figure but descends into a figure's
     * children. Use this method to implement <i>click-through</i>
     * hit detection, that is, you want to detect the inner most
     * figure containing the given point.
     */
    public Figure findFigureInside(int x, int y) {
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure().findFigureInside(x, y);
            if (figure != null)
                return figure;
        }
        return null;
    }

    /**
     * Finds a connection in this. Descends into the
     * children. 
     */
    public Figure findConnection(int x, int y) {
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure().findConnection(x, y);
            if (figure != null)
                return figure;
        }
        return null;
    }

    /**
     * Finds a figure but descends into a figure's
     * children. It supresses the passed
     * in figure. Use this method to ignore a figure
     * that is temporarily inserted into the drawing.
     */
    public Figure findFigureInsideWithout(int x, int y, Figure without) {
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            if (figure != without) {
                Figure found = figure.findFigureInside(x, y);
                if (found != null)
                    return found;
            }
        }
        return null;
    }

    /**
     * Checks if the composite figure has the argument as one of
     * its children.
     */
    public boolean includes(Figure figure) {
        if (super.includes(figure))
            return true;

        FigureEnumeration k = figures();
        while (k.hasMoreElements()) {
            Figure f = k.nextFigure();
            if (f.includes(figure))
                return true;
        }
        return false;
    }

    /**
     * Propagate beginning of drag to subfigures.
     */
    public void beginDrag(MouseEvent e) {
     super.beginDrag(e);
     FigureEnumeration figure = figures();
      while (figure.hasMoreElements())
        figure.nextFigure().beginDrag(e);
    }

    /**
     * Propagate ending of drag to subfigures.
     */
    public void endDrag(MouseEvent e) {
     super.endDrag(e);
     FigureEnumeration figure = figures();
      while (figure.hasMoreElements())
        figure.nextFigure().endDrag(e);
    }

    /**
     * Moves all the given figures by x and y. Doesn't announce
     * any changes. Subclassers override
     * basicMoveBy. Clients usually call moveBy.
     * @see moveBy
     */
    public void basicMoveBy(int x, int y) {
        FigureEnumeration k = figures();
        while (k.hasMoreElements())
            k.nextFigure().moveBy(x,y);
    }

    /**
     * Changes the display box of a composite figure. Clients usually
     * call this method. It changes the display box
     * and announces the corresponding change. Marks the layout
     * invalid (dirty).
     * @param origin the new origin
     * @param corner the new corner
     * @see displayBox
     */
    public void displayBox(Point origin, Point corner) {
    	//	System.err.println("%%%% begin displayBox "+this);
        if (parent() != null)
          if (!((CompositeFigure)parent()).acceptSubFigureChange(origin,corner))
            return;
        willChange();
        basicDisplayBox(origin, corner);
        markDirty(); //invalidate the layout and announce the change
     	//	System.err.println("%%%% end displayBox "+this);
   }
    
    /**
     * Releases the figure and all its children.
     */
    public void release() {
        super.release();
        FigureEnumeration k = figures();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            figure.release();
        }
    }

    /**
     * Invalidates the figure. This method informs the listeners
     * that the figure's current display box is invalid and should be
     * refreshed. The default for all composites is to add the subfigures'
     * display boxes to the rectangle returned by displayBox(). 
     */
    public void invalidate() {
    		Rectangle r = displayBox();
    	  if (fListener != null && r != null) {   
            //System.err.println("------ "+this);
            //System.err.println("11### "+r);
            FigureEnumeration k = figures();
            while (k.hasMoreElements())   {
              Figure f = k.nextFigure();
              //r.add(f.invalidatedBox());
              r.add(f.displayBox());
              //System.err.println("f: "+f);
            }
            r.grow(Handle.HANDLESIZE, Handle.HANDLESIZE);
            //System.err.println("22### "+r);
            fListener.figureInvalidated(new FigureChangeEvent(this, r));
        }
    }


    /**
     * Propagates the figureInvalidated event to my listener.
     * @see FigureChangeListener
     */
    public void figureInvalidated(FigureChangeEvent e) {
        if (listener() != null)
            listener().figureInvalidated(e);
    }

    /**
     * Propagates the removeFromDrawing request up to the container.
     * @see FigureChangeListener
     */
    public void figureRequestRemove(FigureChangeEvent e) {
        if (listener() != null)
            listener().figureRequestRemove(new FigureChangeEvent(this));
    }

    /**
     * Propagates the requestUpdate request up to the container.
     * @see FigureChangeListener
     */
    public void figureRequestUpdate(FigureChangeEvent e) {
        if (listener() != null)
            listener().figureRequestUpdate(e);
    }

    public void figureChanged(FigureChangeEvent e) {
    }

    public void figureRemoved(FigureChangeEvent e) {
    // ??	markDirty();
    }

	/**
	 * The standard implementation of the validation algorithm.
	 * <ol>
	 *	<li> for each subfigure <em>f</em>: if <em>f</em> is not valid,
	 *       validate <em>f</em>.
	 * 	<li> lay out this composite
	 *	<li> set the validation flag to true so that isValid method returns
	 *   		 true
	 * </ol> 
	 */
	public void validate() {
	//	System.err.println("/// validating "+this);
		doLayout();
		FigureEnumeration figures = figures();
		while (figures.hasMoreElements()) {
			Figure figure = figures.nextFigure();
			if (!figure.isValid())
				figure.validate();
		}
		//doLayout();
		markValid(); 
	}
		
   /**
   * Calls the layout manager to lay out this composite.
   */
  public void doLayout() {	
  	//System.err.println("!!! lay out "+this);
  	getLayout().layoutContainer(this);
  }

 /**
   *  Returns the validation status flag.
   */
  public boolean isValid() {
  	return fIsValid;
  }
  
  /**
   * Set the validation status of this composite to false.
   */
  public void markDirty() {
  	//System.err.println("##### dirty "+this);
  	fIsValid = false;
  	getLayout().invalidateLayout();
  	changed();
  }

  
  /**
   * Set the validation status of this composite to false.
   */
  public void markDirty(Figure figure) {
  //	System.err.println("##### dirty "+this+" with "+figure);
  	fIsValid = false;
  	getLayout().invalidateLayout(figure);
  	changed();
  }

  /**
   * Set the validation status of this composite to true.
   */
  public void markValid() {
  //	System.err.println("----- valid "+this);
  	fIsValid = true;
  }
  	
    /**
     * Writes the contained figures to the StorableOutput.
     */
    public void write(StorableOutput dw) {
        super.write(dw);
        dw.writeInt(fFigures.size());
        Enumeration k = fFigures.elements();
        while (k.hasMoreElements())
            dw.writeStorable((Storable) k.nextElement());
    }

    /**
     * Reads the contained figures from StorableInput.
     */
    public void read(StorableInput dr) throws IOException {
        super.read(dr);
        int size = dr.readInt();
        fFigures = new Vector(size);
        for (int i=0; i<size; i++)
            add((Figure)dr.readStorable());
    }

    private void readObject(ObjectInputStream s)
        throws ClassNotFoundException, IOException {

        s.defaultReadObject();

        FigureEnumeration k = figures();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            figure.addToContainer(this);
        }
    }
}