

package CH.ifa.draw.standard;

import CH.ifa.draw.framework.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.Geom;
import java.awt.*;
import java.io.*;

/**
 * Basic implementation of a locator for locations at start-left, start-right,
 * middle-left, middle-right, end-left, and end-right of a line segment.
 * The class provides factory methods for creating instances of LabelSlotLocators
 * that perform the computations for each logical position.
 */
public class SlotLocator extends SegmentLocator {

    protected LabelSlotLocator fSlot;

    /**
     * N/A - do nothing.
     */
    protected void updateSegment(PolyLineFigure pLine) {
    }

    /**
     * Set the segment of this.
     */
    public void setSegment(int segment) {
      fSegment = segment;
    }
    /** 
     * @deprecated 
     */
    public SlotLocator(LabelledConnection conn, Figure attachment) {
      super(conn,attachment);
      fSlot = getLabelSlot(conn,fSegment,attachment);
    }

    /** 
     * @deprecated 
     */
    public SlotLocator(LabelledConnection conn, Figure attachment,int segment) {
      super(conn,attachment,segment);
      fSlot = getLabelSlot(conn,fSegment,attachment);
    }

    /**
     * Use this to create the locator.
     */
    public SlotLocator(LabelledConnection conn, Figure attachment,int segment,
                       LabelSlotLocator slotLocator) {
      super(conn,attachment,segment);
      fSlot = slotLocator;
    }

    /**
     * Find the location. Find first a point on the line segment and
     * then the displacement for the centroid of the located figure.
     */
    protected Point findLocation(Point p1,Point p2) {
      Point place = fSlot.locate(p1,p2);
      Point dXY = fSlot.displace(p1,p2);
      return new Point(place.x + dXY.x,place.y + dXY.y);
    }

    /**
     * @deprecated
     */
    protected LabelSlotLocator getLabelSlot(LabelledConnection conn, int segment,
                                            Figure attachment) {
      Point p1 = ((PolyLineFigure)conn).pointAt(fSegment);
      Point p2 = ((PolyLineFigure)conn).pointAt(fSegment+1);
      Point middle = new Point(p2.x - (p2.x-p1.x)/2,p2.y - (p2.y-p1.y)/2);
      Point att = attachment.center();
      double angle =  Math.atan2(p2.x-p1.x,p2.y-p1.y);
     // System.err.println("att.y: "+att.y+" middle.y: "+middle.y+" p1.y: "+p1.y+" p2.y: "+p2.y);
      if (Geom.length(att.x,att.y,p1.x,p1.y) < Geom.length(att.x,att.y,p2.x,p2.y)) {
        if (Geom.length(att.x,att.y,middle.x,middle.y) < Geom.length(att.x,att.y,p1.x,p1.y)) {
          return new CenterSlot(
            conn,attachment,angle > Math.atan2(att.x-p1.x,att.y-p1.y)
          );
        } else
          return new BeginSlot(conn,attachment,att.y <= p1.y);
      } else {
        if (Geom.length(att.x,att.y,middle.x,middle.y) < Geom.length(att.x,att.y,p2.x,p2.y))
          return new CenterSlot(conn,attachment,att.y <= middle.y);
        else
          return new EndSlot(conn,attachment,att.y <= p2.y);
      }
    }

    /**
     * A Factory method to create a locator for the center of a segment.
     */
    public static LabelSlotLocator
    getCenterSlotLocator(LabelledConnection connection,Figure label,boolean above) {
      return new CenterSlot(connection,label,above);
    }

    /**
     * A Factory method to create a locator for the start of a segment.
     */
    public static LabelSlotLocator
    getBeginSlotLocator(LabelledConnection connection,Figure label,boolean above) {
      return new BeginSlot(connection,label,above);
    }

    /**
     * A Factory method to create a locator for the end of a segment.
     */
    public static LabelSlotLocator
    getEndSlotLocator(LabelledConnection connection,Figure label,boolean above) {
      return new EndSlot(connection,label,above);
    }
}

/**
 * Core implementation shared by concrete LabelSlotLocators.
 */
abstract class AbstractSlot implements LabelSlotLocator,Serializable {

	/** Flag for a position above/below (left/right) the connection. */
  protected boolean fIsAbove = true;

	/** The connection. */
  protected LabelledConnection fConnection;
  
  /** The located figure. */
  protected Figure fLabel;

  public AbstractSlot(LabelledConnection connection,Figure label,boolean above) {
    fConnection = connection;
    fLabel = label;
    fIsAbove = above;
   // System.err.println(this+" above: "+above);
  }

  /*protected double getKRatio(Point p1,Point p2) {
    double kRatio;
    if (p2.y == p1.y)
      kRatio = 0.0;
    else
      if (p2.x != p1.x)
        kRatio = ((double)(p2.y - p1.y)) / ((double)(p2.x - p1.x));
      else
        if (p2.y > p1.y )
          kRatio = Double.POSITIVE_INFINITY;
        else
          kRatio = Double.NEGATIVE_INFINITY;
    return kRatio;
  }*/

	/** 
	 * @deprecated 
	 */
  protected Point getDisplacement(int dX,int dY) {
    if (fIsAbove)
      return new Point(dX,dY);
    else
      return new Point(-dX,-dY);
  }
}

/**
 * A locator for positioning the located figure's centroid on a line
 * perpendicular to a segment of a connection and intersecting the segment
 * at half way. The figure is placed as close to the segment as possible
 * either above or below (left or right looking from the first point of
 * the segment) so that the display box of the figure and the segment line
 * do not intersect. A margin of 3 points is kept, however.
 */
class CenterSlot extends AbstractSlot {

  public CenterSlot(LabelledConnection connection,Figure label) {
    this(connection,label,true);
  }

   public CenterSlot(LabelledConnection connection,Figure label,boolean above) {
    super(connection,label,above);
    //System.err.println("created "+this);
  }

	/** 
	 * Compute the location for the line segment from p1 to p2.
	 */
    public Point displace(Point p1,Point p2) {
      int dX = 0,dY = 0;
      double a, b, c, d, e, f;
      double alpha = Math.atan2(p2.y-p1.y,p2.x-p1.x);
      double w = fLabel.displayBox().width + 6; // leave some space around
      double h = fLabel.displayBox().height + 6;
      //double kRatio = getKRatio(p1,p2);
      if (p2.x >= p1.x && p2.y >= p1.y) {
        double co = Math.cos(alpha);
        double ta = Math.tan(alpha);
        b = h/2 * ta;
        a = w/2 - b;
        c = a * co;
        d = c * Math.sin(alpha);
        e = d * ta;
        if (fIsAbove) {
          dX =  (int)Math.floor(b + e);
          dY = -(int)Math.floor(d + h/2);
        } else {
          dX = -(int)Math.floor(b + e);
          dY =  (int)Math.floor(d + h/2);
        }
      } else if (p2.x < p1.x && p2.y >= p1.y) {
        alpha = Math.PI - alpha;
        double co = Math.cos(alpha);
        double ta = Math.tan(alpha);
        b = h/2 * ta;
        a = w/2 - b;
        c = a * co;
        d = c * Math.sin(alpha);
        e = d * ta;
        if (fIsAbove) {
          dX = (int)Math.floor(b + e);
          dY = (int)Math.floor(d + h/2);
        } else {
          dX = -(int)Math.floor(b + e);
          dY = -(int)Math.floor(d + h/2);
        }
      } else if (p2.x < p1.x && p2.y < p1.y) {
        alpha = Math.PI + alpha;
        double co = Math.cos(alpha);
        double ta = Math.tan(alpha);
        b = h/2 * ta;
        a = w/2 - b;
        c = a * co;
        d = c * Math.sin(alpha);
        e = d * ta;
        if (fIsAbove) {
          dX = -(int)Math.floor(b + e);
          dY =  (int)Math.floor(d + h/2);
        } else {
          dX =  (int)Math.floor(b + e);
          dY = -(int)Math.floor(d + h/2);
        }
      } else if (p2.x >= p1.x && p2.y < p1.y) {
        alpha = - alpha;
        double co = Math.cos(alpha);
        double ta = Math.tan(alpha);
        b = h/2 * ta;
        a = w/2 - b;
        c = a * co;
        d = c * Math.sin(alpha);
        e = d * ta;
        if (fIsAbove) {
          dX = -(int)Math.floor(b + e);
          dY = -(int)Math.floor(d + h/2);
        } else {
          dX =  (int)Math.floor(b + e);
          dY =  (int)Math.floor(d + h/2);
        }
      }
      return new Point(dX,dY);
    }

  public Point locate(Point p1,Point p2) {
      return new Point(
        p1.x + (p2.x - p1.x)/2, p1.y + (p2.y - p1.y)/2
      );

  }
}

/**
 * A locator for positioning the upper/left or lower/left corner of 
 * the display box of the located figure on a line
 * perpendicular to a segment of a connection and intersecting the segment
 * at the first point. The figure is placed as close to the segment as possible
 * either above or below (left or right looking from the first point of
 * the segment) so that the display box of the figure and the segment line
 * do not intersect. A margin of 3 points is kept, however.
 */

class BeginSlot extends AbstractSlot {

  public BeginSlot(LabelledConnection connection,Figure label) {
    this(connection,label,true);
  }

   public BeginSlot(LabelledConnection connection,Figure label,boolean above) {
    super(connection,label,above);
  }

  /** 
	 * Compute the location for the line segment from p1 to p2.
	 */
  public Point displace(Point p1,Point p2) {
      int dX = 0,dY = 0;
      double a, b,c, e, f;
      double alpha = Math.atan2(p2.y-p1.y,p2.x-p1.x);
      int w = fLabel.displayBox().width + 6; // leave some space around
      int h = fLabel.displayBox().height + 6;
      if (p2.x >= p1.x && p2.y >= p1.y) {
        double theta = Math.PI + alpha; // -PI <= alpha <= 0
        double cot = Math.cos(theta);
        double sit = Math.sin(theta);
        if (fIsAbove) {
          c = h * sit;
          b = c * sit;
          e = c * cot;
          dX = (int)Math.floor(w/2 + e);
          dY = (int)Math.floor(b - h/2);
        } else {
          c = w * cot;
          b = c * sit;
          f = c * cot;
          e = w - f;
          a = h/2 + b;
          dX = (int)Math.floor(w/2 - e);
          dY = (int)Math.floor(a);
        }
      } else if (p2.x < p1.x && p2.y >= p1.y) {
        double sia = Math.sin(-alpha);
        double coa = Math.cos(-alpha);
        if (fIsAbove) {
          c = w * coa;
          b = c * sia;
          f = c * coa;
          e = w - f;
          dX = (int)Math.floor(w/2 - f);
          dY = (int)Math.floor(b + h/2);
        } else {
          c = h * sia;
          b = c * sia;
          e = c * coa;
          dX = (int)Math.floor(-e - w/2);
          dY = (int)Math.floor(b - h/2);
        }
      } else if (p2.x < p1.x && p2.y < p1.y) {
        double sia = Math.sin(alpha);
        double coa = Math.cos(alpha);
        if (fIsAbove) {
          c = h * sia;
          b = c * sia;
          e = c * coa;
          a = h/2 - b;
          dX = (int)Math.floor(-e - w/2);
          dY = (int)Math.floor(a);
        } else {
          c = w * coa;
          f = c * coa;
          e = w - f;
          b = c * sia;
          a = h / 2 + b;
          dX = (int)Math.floor(e - w/2 -3);
          dY = (int)Math.floor(-a);
        }
      } else if (p2.x >= p1.x && p2.y < p1.y) {
        double beta = Math.PI - alpha;
        double sib = Math.sin(beta);
        double cob = Math.cos(beta);
        if (fIsAbove) {
          c = w * cob;
          b = c * sib;
          f = c * cob;
          a = h/2 + b;
          dX = (int)Math.floor(f - w/2);
          dY = (int)Math.floor(-a);
        } else {
          c = h * sib; // Math.min(w * co, d / 3.0);        //??????
          e = c * cob;
          b = c * sib;
          a = h / 2 - b;
          dX = (int)Math.floor(w/2 + e);
          dY = (int)Math.floor(a);
        }
      }
      return new Point(dX,dY);
    }

  public Point locate(Point p1,Point p2) {
      return new Point(
        p1.x, p1.y
      );
  }
}

/**
 * A locator for positioning the upper/right or lower/right corner of 
 * the display box of the located figure on a line
 * perpendicular to a segment of a connection and intersecting the segment
 * at the second point. The figure is placed as close to the segment as possible
 * either above or below (left or right looking from the first point of
 * the segment) so that the display box of the figure and the segment line
 * do not intersect. A margin of 3 points is kept, however.
 */
class EndSlot extends AbstractSlot {

  public EndSlot(LabelledConnection connection,Figure label) {
    this(connection,label,true);
  }

  public EndSlot(LabelledConnection connection,Figure label,boolean above) {
    super(connection,label,above);
  }

	/** 
	 * Compute the location for the line segment from p1 to p2.
	 */
    public Point displace(Point p1,Point p2) {
      int dX = 0,dY = 0;
      double a, b,c, e, f;
      double alpha = Math.atan2(p2.y-p1.y,p2.x-p1.x);
      int w = fLabel.displayBox().width + 6; // leave some space around
      int h = fLabel.displayBox().height + 6;
      if (p2.x >= p1.x && p2.y >= p1.y) {
        double sia = Math.sin(alpha);
        double coa = Math.cos(alpha);
        if (fIsAbove) {
          c = w * coa;
          f = c * coa;
          e = w - f;
          b = c * sia;
          a = h / 2 + b;
          dX = (int)Math.floor(e - w/2 -3);
          dY = (int)Math.floor(-a);
        } else {
          c = h * sia;
          b = c * sia;
          e = c * coa;
          a = h/2 - b;
          dX = (int)Math.floor(-e - w/2);
          dY = (int)Math.floor(a);
        }
      } else if (p2.x < p1.x && p2.y >= p1.y) {
        double beta = Math.PI - alpha;
        double sib = Math.sin(beta);
        double cob = Math.cos(beta);
        if (fIsAbove) {
          c = h * sib; // Math.min(w * co, d / 3.0);        //??????
          e = c * cob;
          b = c * sib;
          a = h / 2 - b;
          dX = (int)Math.floor(w/2 + e);
          dY = (int)Math.floor(a);
        } else {
          c = w * cob;
          b = c * sib;
          f = c * cob;
          a = h/2 + b;
          dX = (int)Math.floor(f - w/2);
          dY = (int)Math.floor(-a);
        }
      } else if (p2.x < p1.x && p2.y < p1.y) {
        double theta = Math.PI + alpha; // -PI <= alpha <= 0
        double cot = Math.cos(theta);
        double sit = Math.sin(theta);
        if (fIsAbove) {
          c = w * cot;
          b = c * sit;
          f = c * cot;
          e = w - f;
          a = h/2 + b;
          dX = (int)Math.floor(w/2 - e);
          dY = (int)Math.floor(a);
        } else {
          c = h * sit;
          b = c * sit;
          e = c * cot;
          dX = (int)Math.floor(w/2 + e);
          dY = (int)Math.floor(b - h/2);
        }
      } else if (p2.x >= p1.x && p2.y < p1.y) {
        double sia = Math.sin(-alpha);
        double coa = Math.cos(-alpha);
        if (fIsAbove) {
          c = h * sia;
          b = c * sia;
          e = c * coa;
          dX = (int)Math.floor(-e - w/2);
          dY = (int)Math.floor(b - h/2);
        } else {
          c = w * coa;
          b = c * sia;
          f = c * coa;
          e = w - f;
          dX = (int)Math.floor(w/2 - f);
          dY = (int)Math.floor(b + h/2);
        }
      }
      return new Point(dX,dY);
    }

  public Point locate(Point p1,Point p2) {
    return new Point(p2.x,p2.y);
  }
}