/*
 * @(#)StandardDrawing.java 5.1
 *
 */

package CH.ifa.draw.standard;

import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import java.awt.*;
import java.util.*;
import java.io.*;

/**
 * The standard implementation of the Drawing interface.
 *
 * @see Drawing
 */

public class StandardDrawing extends CompositeFigure implements Drawing {


    /**
     * the registered listeners
     */
    private transient Vector              fListeners;

    /**
     * boolean that serves as a condition variable
     * to lock the access to the drawing.
     * The lock is recursive and we keep track of the current
     * lock holder.
     */
    private transient Thread    fDrawingLockHolder = null;

    /*
     * Serialization support
     */
    private static final long serialVersionUID = -2602151437447962046L;
    private int drawingSerializedDataVersion = 1;

    /**
     * Constructs the Drawing.
     */
    public StandardDrawing() {
        super();
        fListeners = new Vector(2);
    }

    /**
     * Brings a figure to the front.
     */
    public synchronized void bringToFront(Figure figure) {
        if (fFigures.contains(figure)) {
            //System.err.println(">>>StandardD#bringToFront "+figure);
            fFigures.removeElement(figure);
            fFigures.addElement(figure);
            figure.broughtToFront();
            figure.changed();
            // no need to propagate event further - this is the root
            // of figure hierarchy
        }
    }

    /**
     * Any figure can be dropped onto a drawing.
     */

    public boolean accept(Figure figure) {
      return true;
    }

    /**
     * Add the dragged figure to this.
     */
     public Figure drop(Figure dragTarget) {
        if (accept(dragTarget))
          ParentChangeEvent.changeParent(dragTarget,this);
        return dragTarget;
     }

     /**
      * CompositeFigure serves as a base abstraction for GroupFigure.
      * Unfortunately Drawing is not a GroupFigure - it is more like
      * PanelComposite without a frame. This is why we can't use
      * the findFigure() in CompositeFigure for hit detection. Hit
      * detection should not descend into the children of a GroupFigure
      * but it should descend into the children of PanelComposites.  
      */
     public Figure findSelectable(int x, int y) {
      //System.err.println("findSelectable"+this);

        Figure figure = findConnection(x,y);
        if (figure != null)
          return figure;
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            figure = k.nextFigure();
            Figure selectable = null;
            //if (figure.containsPoint(x, y)) {
            selectable = figure.findSelectable(x,y);
            if (selectable != null)
              return selectable;
            //}
        }
        return null;
     }

     public Figure findCtrlSelectable(int x, int y) {
      	//System.err.println("findSelectable"+this);
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            Figure ctrlSelectable = null;
            //if (figure.containsPoint(x, y)) {
            ctrlSelectable = figure.findCtrlSelectable(x,y);
            if (ctrlSelectable != null)
              return ctrlSelectable;
            //}
        }
        return null;
     }

     /**
      * Find the figure that can be connected at the given point.
      * Return null, if no figure is found.
      */
    public Figure findConnectable(int x, int y,Figure connection) {
      FigureEnumeration k = figuresReverse();
      while (k.hasMoreElements()) {
        Figure figure = (k.nextFigure()).findConnectable(x,y,connection);
        if (figure != null)
          return figure;
      }
      return null;
    }
     /**
      * Find a figure at the given point that can accept the
      * dragTarget. Return this, if no other figure is found (can always
      * add the figure in the drawing).
      */
     public Figure findDropTarget(int x, int y, Figure dragTarget) {
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            //System.err.println("@@ figure: "+figure);
            //if (figure != dragTarget && figure.containsPoint(x,y)) {
            // don't drop a figure on itself
              //System.err.println("@@ ok");
            Figure found = figure.findDropTarget(x, y, dragTarget);
            if (found != null)
              return found;
            //}
        }
        return this; // can always drop on this
     }

    /**
     * Adds a listener for this drawing.
     */
    public void addDrawingChangeListener(DrawingChangeListener listener) {
        fListeners.addElement(listener);
    }

    /**
     * Removes a listener from this drawing.
     */
    public void removeDrawingChangeListener(DrawingChangeListener listener) {
        fListeners.removeElement(listener);
    }

    /**
     * Adds a listener for this drawing.
     */
    public Enumeration drawingChangeListeners() {
        return fListeners.elements();
    }

    /**
     * Removes the figure from the drawing and releases it.
     */

    public synchronized Figure remove(Figure figure) {
        // ensure that we remove the top level figure in a drawing
        //System.err.println("removing: "+figure);
        if (figure.listener() != null) {
        	//System.err.println("figure.listener(): "+figure.listener());
            figure.listener().figureRequestRemove(new FigureChangeEvent(figure, null));
            return figure;
        }
        return null;
    }


    /**
     * Handles a removeFromDrawing request that
     * is passed up the figure container hierarchy.
     * @see FigureChangeListener
     */
    public void figureRequestRemove(FigureChangeEvent e) {
        Figure figure = e.getFigure();
        if (fFigures.contains(figure)) {
            fFigures.removeElement(figure);
            figure.removeFromContainer(this);   // will invalidate figure
            figure.release();
        } else
            System.out.println("Attempt to remove non-existing figure");
    }

    /**
     * Remove a top-level figure from this (the drawing), but  do not release it.
     * First orphan it, and then remove this from the containers of the figure.
     */
    public Figure removeChild(Figure childFigure) {
 			orphan(childFigure);
 			childFigure.removeFromContainer(this);
      return childFigure;
    }
    
  /**
   * Cause immediate validation of the layout of the drawing???
   * /
  public void markDirty() {
  	validate();
  } */

    public void figureChanged(FigureChangeEvent e) {
    	validate();
    }
    
    /**
     * Invalidates a rectangle and merges it with the
     * existing damaged area.
     * @see FigureChangeListener
     */
    public void figureInvalidated(FigureChangeEvent e) {
        if (fListeners != null) {
            for (int i = 0; i < fListeners.size(); i++) {
                DrawingChangeListener l = (DrawingChangeListener)fListeners.elementAt(i);
                l.drawingInvalidated(new DrawingChangeEvent(this, e.getInvalidatedRectangle()));
            }
        }
    }

    /**
     * Forces an update
     */
    public void figureRequestUpdate(FigureChangeEvent e) {
        if (fListeners != null) {
            for (int i = 0; i < fListeners.size(); i++) {
                DrawingChangeListener l = (DrawingChangeListener)fListeners.elementAt(i);
                l.drawingRequestUpdate(new DrawingChangeEvent(this, null));
            }
        }
    }

    /**
     * Return's the figure's handles. This is only used when a drawing
     * is nested inside another drawing.
     */
    public Vector handles() {
        Vector handles = new Vector();
        handles.addElement(new NullHandle(this, RelativeLocator.northWest()));
        handles.addElement(new NullHandle(this, RelativeLocator.northEast()));
        handles.addElement(new NullHandle(this, RelativeLocator.southWest()));
        handles.addElement(new NullHandle(this, RelativeLocator.southEast()));
        return handles;
    }

    /**
     * Gets the display box. This is the union of all figures.
     */
    public Rectangle displayBox() {
        if (fFigures.size() > 0) {
            FigureEnumeration k = figures();

            Rectangle r = k.nextFigure().displayBox();

            while (k.hasMoreElements())
                r.add(k.nextFigure().displayBox());
            return r;
        }
        return new Rectangle(0, 0, 0, 0);
    }

    public void basicDisplayBox(Point p1, Point p2) {
    }

    /**
     * Acquires the drawing lock.
     */
    public synchronized void lock() {
        // recursive lock
        Thread current = Thread.currentThread();
        if (fDrawingLockHolder == current)
            return;
        while (fDrawingLockHolder != null) {
            try { wait(); } catch (InterruptedException ex) { }
        }
        fDrawingLockHolder = current;
    }

    /**
     * Releases the drawing lock.
     */
    public synchronized void unlock() {
        if (fDrawingLockHolder != null) {
            fDrawingLockHolder = null;
            notifyAll();
        }
    }

    private void readObject(ObjectInputStream s)
        throws ClassNotFoundException, IOException {

        s.defaultReadObject();

        fListeners = new Vector(2);
        FigureEnumeration k = figures();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            figure.addToContainer(this);
            figure.setParent(this);
        }
    }
}