package relap.Editor;

import relap.Analyzer.*;
import relap.LanguageModel.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.figures.*;
import java.io.*;
import java.util.*;
import java.awt.event.*;
import java.awt.*;
import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.*;
import com.objectspace.jgl.predicates.*;

public abstract class AbstractAnalyzer 
	implements Analyzer {
	
	protected HashMap fImplicitPredicateDescriptions;
	protected IndexedMDSet fParserInput;
	protected SyntaxTreeNode fSemantics;
	protected DrawingEditor fEditor;	
	protected ARGRecognizer fParser;	
	protected Vector fObservers;
	protected HashSet currentSelection;
	protected boolean fWriteLog;

	public AbstractAnalyzer(DrawingEditor editor) {
		fEditor = editor;
		currentSelection = new HashSet();
		fWriteLog = false;
	}
		
	public void register(RelapObserver o) {
		if (fObservers==null)
			fObservers = new Vector();
		if (!fObservers.contains(o))
			fObservers.addElement(o);
	}
	
	public void unregister(RelapObserver o) {
		if (fObservers!=null && fObservers.contains(o))
			fObservers.removeElement(o);
	}
	
	protected void notify(RelapEvent e) {
		if (fObservers != null) {
			Enumeration o = fObservers.elements();
			while (o.hasMoreElements())
				((RelapObserver)o.nextElement()).update(e);
		}
	}
	
	public ARGRecognizer getParser() {
		if (fParser == null)
			fParser = getDefaultParser();
		return fParser;
	}
	
	protected ARGRecognizer getDefaultParser() {
		return new FixedStartParser();
	}
	 	
	abstract protected String getGrammar();
	
	abstract protected ObjectToVisitorMap getVisitorMap();
	
	abstract protected ObjectToSymbolMap getSymbolMap();
	
	protected boolean loggingOn() {
		return fWriteLog;
	}

	public void setLogging(boolean mode) {
		fWriteLog = mode;
	}

	public Drawing getDrawing() {
		return fEditor.drawing();
	}
	
	public void setEditor(DrawingEditor d) {
		fEditor = d;
	}
	
	public DrawingEditor getEditor() {
		return fEditor;
	}

	public SyntaxTreeNode getSemantics() {
		return fSemantics;
	}
	
	public void setSemantics(SyntaxTreeNode node) {
		fSemantics = node;	
	}

	public SemanticProcessor getSemanticProcessor() {
		return new TreePrinter();
	}

	protected void setPredicateDescriptions() {
		fImplicitPredicateDescriptions = new HashMap();	
	}
	
	public String describePredicate(String predicateName) {
		return (String)fImplicitPredicateDescriptions.get(predicateName);
	}
	
	public String describeMissingRelationTuple(String relationName,FailureData fData) {
		ExpanderRelation relation = 
			(ExpanderRelation)fParserInput.getRelation(relationName);
		if (relation != null) 
			return relation.describeFailure(fData,fParserInput);
		return relationName+" -> no such relation";
	}

	public void visitDrawing(Object visitorState,ObjectToVisitorMap visitors)
		throws VisitorException  {
		visitors.mapObjectToVisitor(getDrawing()).visit(getDrawing(),visitorState);
	}
	
	abstract protected void initRelations(IndexedMDSet input);	
	
	abstract protected Object getStartObject(IndexedMDSet input,ObjectToSymbolMap map);	

	protected InputGenerationState initAnalysisInput(ObjectToVisitorMap generators,
									   ObjectToSymbolMap symbols) {
		return new InputGenerationState(generators,symbols);
	}
		
	public void analyzeDrawing() {
		analyzeDrawing(getGrammar(),getVisitorMap(),getSymbolMap(),loggingOn());
	}
	
	public void analyzeDrawing(String grammar,ObjectToVisitorMap generators,
	                           ObjectToSymbolMap symbols,boolean writeLog) {
		RelationalGrammarImplementation g; 
		ARGRecognizer parser = getParser();
		PrintWriter log = null;
	 	try {
			notify(new StartEvent(this));
	 		g = (RelationalGrammarImplementation)
	 				Class.forName(grammar).newInstance();
	 		//System.err.println("logs"+File.pathSeparator+grammar);
	 		if (writeLog) {
	 			String separator = System.getProperty("file.separator");
	 			File log_dir = new File("logs"+separator+grammar);
	 			if (!log_dir.exists()) 
	 				if (!log_dir.mkdirs())
	 					System.err.println("couldn't create dir "+log_dir.getAbsolutePath());
	 			String fileName = 
	 				"logs"+separator+grammar+separator+
	 					(new Date()).toString()
	 						.replace(' ','_')
	 						.replace(':','_')
	 						.replace('+','_')+".rlog";
	 			File f = new File(fileName);
	 			File ff = new File("puppu");
	 			System.err.println("### "+fileName);
				log = new PrintWriter(new FileOutputStream(f));
			}
			g.setPredicateEvaluator(this);
			g.buildGrammar();
			if (log != null)
				log.println(g.toString());
			//System.err.println(g.toString());
			InputGenerationState genState = 
				initAnalysisInput(generators,symbols);
			setPredicateDescriptions(); 
			initRelations(genState.input);
			visitDrawing(genState,genState.generators);
			genState.input.updateObjects();
			genState.input.setObjectClassToTerminalMap(symbols);
			fParserInput = genState.input;
			//System.err.println(genState.input);
			Object start = getStartObject(genState.input,genState.symbols);
			if (start == null) {
				if (log != null) 
					log.close();
			}
			else {
				preParseActions();
				ProcessingResult finalResult = null;
				ParseResult firstResult = 
					parser.parse(genState.input,g,start,log,1000000);
				if (firstResult.success())
					finalResult = postParseActions(firstResult);
				else
					finalResult = firstResult;
				if (log != null) {
					finalResult.reportOn(log);
					log.close();
				}
				System.err.println("done");
				// Report parse result 
				notify(new StopEvent(this,finalResult));
			}
		}
		catch(InvalidGrammarException e) {
			notify(new AnalysisExceptionEvent(this,e));
		}	
		catch(AnalysisException e) {
			notify(new AnalysisExceptionEvent(this,e));
		}	
		catch(IOException e) {
			notify(new AnalysisExceptionEvent(this,e));
		}	
		catch(SecurityException e) {
			notify(new AnalysisExceptionEvent(this,e));
		}	
		catch (ClassNotFoundException e) {
			notify(new AnalysisExceptionEvent(this,e));
		} 
		catch (InstantiationException e) {
			notify(new AnalysisExceptionEvent(this,e));
		} 
		catch (IllegalAccessException e) {
			notify(new AnalysisExceptionEvent(this,e));
		} 
		catch (VisitorException e) {
			notify(new AnalysisExceptionEvent(this,e));
		} 
	}
	
	protected void preParseActions() throws AnalysisException {
		// hook
	}

	protected ProcessingResult postParseActions(ParseResult result) 
		throws AnalysisException {
		return result;
	}
		
	
	public void setSelection(com.objectspace.jgl.Container objects) {
		clearSelection();
		Copying.copy(objects,currentSelection);
		highlightSelection();
	}
	
	public void highlightSelection() {
		Enumeration object = currentSelection.elements();
		while (object.hasMoreElements()) {
			Figure figure =  (Figure)object.nextElement();
		  figure.setHighlighted(true);
			fEditor.view().checkDamage();		
		}
	}
	
	public void clearSelection() {
		Enumeration object = currentSelection.elements();
		while (object.hasMoreElements()) {
			Figure figure =  (Figure)object.nextElement();
			//System.err.println(">> in clear: "+figure);
		  figure.setHighlighted(false);
			fEditor.view().checkDamage();
		}
		currentSelection.clear();
	}
	
}